@@ -75,7 +75,7 @@ static void hash_binary(char *hex, unsigned char **password, unsigned int *seed)
 	if (strlen(hex) != 40)
 		fatal("Incorrect password format %s\n", hex);
 
-	md5 = malloc(20);
+	md5 = bip_malloc(20);
 	for (i = 0; i < 20; i++) {
 		sscanf(hex + 2 * i, "%02x", &buf);
 		md5[i] = buf;
@@ -133,7 +133,7 @@ void conf_die(bip_t *bip, char *fmt, ...)
 	va_list ap;
 	int size = ERRBUFSZ;
 	int n;
-	char *error = malloc(size);
+	char *error = bip_malloc(size);
 
 	for (;;) {
 		va_start(ap, fmt);
@@ -147,7 +147,7 @@ void conf_die(bip_t *bip, char *fmt, ...)
 			size = n + 1;
 		else
 			size *= 2;
-		error = realloc(error, size);
+		error = bip_realloc(error, size);
 	}
 	va_start(ap, fmt);
 	_mylog(LOG_ERROR, fmt, ap);
@@ -195,7 +195,10 @@ int do_pid_stuff(void)
 	char hname[1024];
 	char longpath[1024];
 	FILE *f;
+	int fd;
+
 try_again:
+	fd = -1;
 	f = fopen(conf_pid_file, "r");
 	if (f)
 		goto pid_is_there;
@@ -203,7 +206,6 @@ try_again:
 		fatal("%s %s", "gethostname", strerror(errno));
 	snprintf(longpath, 1023, "%s.%s.%ld", conf_pid_file, hname,
 			(long unsigned int)getpid());
-	int fd;
 	if ((fd = open(longpath, O_CREAT|O_WRONLY, S_IWUSR|S_IRUSR)) == -1)
 		fatal("Cannot write to PID file (%s) %s", longpath,
 				strerror(errno));
@@ -222,8 +224,11 @@ pid_is_there:
 	{
 		pid_t pid;
 		long unsigned int p;
+		if (fd != -1)
+			close(fd);
 		if (f) {
 			int c = fscanf(f, "%ld", &p);
+			fclose(f);
 			pid = p;
 			if (c != 1 || p == 0) {
 				mylog(LOG_INFO, "pid file found but invalid "
@@ -240,7 +245,6 @@ pid_is_there:
 		int kr = kill(pid, 0);
 		if (kr == -1 && (errno == ESRCH || errno == EPERM)) {
 			/* that's not bip! */
-			fclose(f);
 			if (unlink(conf_pid_file)) {
 				fatal("Cannot delete pid file '%s', check "
 						"permissions.\n",
@@ -250,7 +254,8 @@ pid_is_there:
 		}
 		if (pid)
 			mylog(LOG_INFO, "pid file found (pid %ld).", pid);
-		mylog(LOG_FATAL, "Another instance of bip is certainly runing.");
+		mylog(LOG_FATAL,
+				"Another instance of bip is certainly runing.");
 		mylog(LOG_FATAL, "If you are sure this is not the case remove"
 					" %s.", conf_pid_file);
 		exit(2);
@@ -341,7 +346,7 @@ static int add_network(bip_t *bip, list_t *data)
 		n->serverv = NULL;
 		n->serverc = 0;
 	} else {
-		n = calloc(sizeof(struct network), 1);
+		n = bip_calloc(sizeof(struct network), 1);
 		hash_insert(&bip->networks, name, n);
 	}
 
@@ -356,7 +361,7 @@ static int add_network(bip_t *bip, list_t *data)
 			break;
 #endif
 		case LEX_SERVER:
-			n->serverv = realloc(n->serverv, (n->serverc + 1)
+			n->serverv = bip_realloc(n->serverv, (n->serverc + 1)
 						* sizeof(struct server));
 			n->serverc++;
 			memset(&n->serverv[n->serverc - 1], 0,
@@ -382,7 +387,7 @@ static int add_network(bip_t *bip, list_t *data)
 	return 1;
 }
 
-void adm_bip_delconn(bip_t *bip, struct link_client *ic, char *conn_name)
+void adm_bip_delconn(bip_t *bip, struct link_client *ic, const char *conn_name)
 {
 	struct user *user = LINK(ic)->user;
 	struct link *l;
@@ -396,8 +401,8 @@ void adm_bip_delconn(bip_t *bip, struct link_client *ic, char *conn_name)
 	link_kill(bip, l);
 }
 
-void adm_bip_addconn(bip_t *bip, struct link_client *ic, char *conn_name,
-		char *network_name)
+void adm_bip_addconn(bip_t *bip, struct link_client *ic, const char *conn_name,
+		const char *network_name)
 {
 	struct user *user = LINK(ic)->user;
 	struct network *network;
@@ -417,7 +422,7 @@ void adm_bip_addconn(bip_t *bip, struct link_client *ic, char *conn_name,
 
 	struct link *l;
 	l = irc_link_new();
-	l->name = strdup(conn_name);
+	l->name = bip_strdup(conn_name);
 	hash_insert(&user->connections, conn_name, l);
 	list_add_last(&bip->link_list, l);
 	l->user = user;
@@ -429,7 +434,7 @@ void adm_bip_addconn(bip_t *bip, struct link_client *ic, char *conn_name,
 #endif
 
 
-#define SCOPY(member) l->member = (LINK(ic)->member ? strdup(LINK(ic)->member) : NULL)
+#define SCOPY(member) l->member = (LINK(ic)->member ? bip_strdup(LINK(ic)->member) : NULL)
 #define ICOPY(member) l->member = LINK(ic)->member
 
 	SCOPY(connect_nick);
@@ -474,7 +479,7 @@ static int add_connection(bip_t *bip, struct user *user, list_t *data)
 #endif
 	} else {
 		l->network = NULL;
-		log_reinit_all(l->log);
+		log_reset_all(l->log);
 	}
 
 	while ((t = list_remove_first(data))) {
@@ -543,12 +548,18 @@ static int add_connection(bip_t *bip, struct user *user, list_t *data)
 			}
 			list_free(t->pdata);
 			break;
+		case LEX_AUTOJOIN_ON_KICK:
+			l->autojoin_on_kick = t->ndata;
+			break;
 		case LEX_FOLLOW_NICK:
 			l->follow_nick = t->ndata;
 			break;
 		case LEX_IGN_FIRST_NICK:
 			l->ignore_first_nick = t->ndata;
 			break;
+		case LEX_IGNORE_CAPAB:
+			l->ignore_server_capab = t->ndata;
+			break;
 		case LEX_AWAY_NICK:
 			MOVE_STRING(l->away_nick, t->pdata);
 			break;
@@ -593,7 +604,7 @@ static int add_connection(bip_t *bip, struct user *user, list_t *data)
 			conf_die(bip, "No nick set and no default nick.");
 			return 0;
 		}
-		l->connect_nick = strdup(user->default_nick);
+		l->connect_nick = bip_strdup(user->default_nick);
 	}
 	if (!l->username) {
 		if (!user->default_username) {
@@ -601,7 +612,7 @@ static int add_connection(bip_t *bip, struct user *user, list_t *data)
 					"username.");
 			return 0;
 		}
-		l->username = strdup(user->default_username);
+		l->username = bip_strdup(user->default_username);
 	}
 	if (!l->realname) {
 		if (!user->default_realname) {
@@ -609,7 +620,7 @@ static int add_connection(bip_t *bip, struct user *user, list_t *data)
 					"realname.");
 			return 0;
 		}
-		l->realname = strdup(user->default_realname);
+		l->realname = bip_strdup(user->default_realname);
 	}
 
 	l->in_use = 1;
@@ -667,7 +678,7 @@ static int add_user(bip_t *bip, list_t *data, struct historical_directives *hds)
 	}
 	u = hash_get(&bip->users, name);
 	if (!u) {
-		u = calloc(sizeof(struct user), 1);
+		u = bip_calloc(sizeof(struct user), 1);
 		hash_insert(&bip->users, name, u);
 		hash_init(&u->connections, HASH_NOCASE);
 		u->admin = 0;
@@ -677,6 +688,7 @@ static int add_user(bip_t *bip, list_t *data, struct historical_directives *hds)
 		u->backlog_lines = DEFAULT_BACKLOG_LINES;
 		u->backlog_no_timestamp = DEFAULT_BACKLOG_NO_TIMESTAMP;
 		u->blreset_on_talk = DEFAULT_BLRESET_ON_TALK;
+		u->blreset_connection = DEFAULT_BLRESET_CONNECTION;
 		u->bip_use_notice = DEFAULT_BIP_USE_NOTICE;
 	} else {
 		FREE(u->name);
@@ -737,6 +749,9 @@ static int add_user(bip_t *bip, list_t *data, struct historical_directives *hds)
 		case LEX_BLRESET_ON_TALK:
 			u->blreset_on_talk = t->ndata;
 			break;
+		case LEX_BLRESET_CONNECTION:
+			u->blreset_connection = t->ndata;
+			break;
 		case LEX_BIP_USE_NOTICE:
 			u->bip_use_notice = t->ndata;
 			break;
@@ -821,6 +836,7 @@ static int validate_config(bip_t *bip)
 						user->name, link->name);
 					link_kill(bip, link);
 					r = 0;
+					continue;
 				}
 			}
 
@@ -833,7 +849,8 @@ static int validate_config(bip_t *bip)
 						"%s: channel must have"
 						"a name.", user->name,
 						link->name);
-					return 0;
+					r = 0;
+					continue;
 				}
 			}
 		}
@@ -844,7 +861,8 @@ static int validate_config(bip_t *bip)
 				"lines to a non-nul value for each user with"
 				"backlog = true. Faulty user is %s",
 				user->name);
-			return 0;
+			r = 0;
+			continue;
 		}
 	}
 
@@ -898,8 +916,8 @@ void sweep(bip_t *bip)
 		if (!l->in_use) {
 			mylog(LOG_INFO, "Administratively killing %s/%s",
 					l->user->name, l->name);
-			link_kill(bip, l);
 			list_remove_if_exists(&bip->conn_list, l);
+			link_kill(bip, l);
 			list_it_remove(&lit);
 		}
 	}
@@ -1147,7 +1165,7 @@ int main(int argc, char **argv)
 	bip_init(&bip);
 	_bip = &bip;
 
-	conf_ip = strdup("0.0.0.0");
+	conf_ip = bip_strdup("0.0.0.0");
 	conf_port = 7778;
 	conf_css = 0;
 
@@ -1160,7 +1178,7 @@ int main(int argc, char **argv)
 	signal(SIGXCPU, rlimit_cpu_reached);
 
 	conf_log_root = NULL;
-	conf_log_format = strdup(DEFAULT_LOG_FORMAT);
+	conf_log_format = bip_strdup(DEFAULT_LOG_FORMAT);
 	conf_log_level = DEFAULT_LOG_LEVEL;
 	conf_daemonize = 1;
 	conf_global_log_file = stderr;
@@ -1172,13 +1190,13 @@ int main(int argc, char **argv)
 	while ((ch = getopt(argc, argv, "hvnf:s:")) != -1) {
 		switch (ch) {
 		case 'f':
-			confpath = strdup(optarg);
+			confpath = bip_strdup(optarg);
 			break;
 		case 'n':
 			conf_daemonize = 0;
 			break;
 		case 's':
-			conf_biphome = strdup(optarg);
+			conf_biphome = bip_strdup(optarg);
 			break;
 		case 'v':
 			version();
@@ -1201,7 +1219,7 @@ int main(int argc, char **argv)
 		return 0;
 	}
 #ifdef HAVE_OIDENTD
-	bip.oidentdpath = malloc(strlen(home) + 1 +
+	bip.oidentdpath = bip_malloc(strlen(home) + 1 +
 			strlen(OIDENTD_FILENAME) + 1);
 	strcpy(bip.oidentdpath, home);
 	strcat(bip.oidentdpath, "/");
@@ -1210,16 +1228,15 @@ int main(int argc, char **argv)
 
 
 	if (!conf_biphome) {
-		conf_biphome = malloc(strlen(home) + strlen("/.bip") + 1);
+		conf_biphome = bip_malloc(strlen(home) + strlen("/.bip") + 1);
 		strcpy(conf_biphome, home);
 		strcat(conf_biphome, "/.bip");
 	}
 
 	if (!confpath) {
-		confpath = malloc(strlen(conf_biphome) + 1 +
+		confpath = bip_malloc(strlen(conf_biphome) + 1 +
 				strlen(S_CONF) + 1);
-		*confpath = 0;
-		strcat(confpath, conf_biphome);
+		strcpy(confpath, conf_biphome);
 		strcat(confpath, "/");
 		strcat(confpath, S_CONF);
 	}
@@ -1234,14 +1251,16 @@ int main(int argc, char **argv)
 
 	if (!conf_log_root) {
 		char *ap = "/logs";
-		conf_log_root = malloc(strlen(conf_biphome) + strlen(ap) + 1);
+		conf_log_root = bip_malloc(strlen(conf_biphome) +
+				strlen(ap) + 1);
 		strcpy(conf_log_root, conf_biphome);
 		strcat(conf_log_root, ap);
 		mylog(LOG_INFO, "Default log root: %s", conf_log_root);
 	}
 	if (!conf_pid_file) {
 		char *pid = "/bip.pid";
-		conf_pid_file = malloc(strlen(conf_biphome) + strlen(pid) + 1);
+		conf_pid_file = bip_malloc(strlen(conf_biphome) +
+				strlen(pid) + 1);
 		strcpy(conf_pid_file, conf_biphome);
 		strcat(conf_pid_file, pid);
 		mylog(LOG_INFO, "Default pid file: %s", conf_pid_file);
@@ -1254,7 +1273,7 @@ int main(int argc, char **argv)
 
 		if (!conf_ssl_certfile) {
 			char *ap = "/bip.pem";
-			conf_ssl_certfile = malloc(strlen(conf_biphome) +
+			conf_ssl_certfile = bip_malloc(strlen(conf_biphome) +
 					strlen(ap) + 1);
 			strcpy(conf_ssl_certfile, conf_biphome);
 			strcat(conf_ssl_certfile, ap);
@@ -1377,7 +1396,7 @@ void adm_print_connection(struct link_client *ic, struct link *lnk,
 noroom: /* that means the line is larger that RET_STR_LEN. We're not likely to
 	   even read such a long line */
 	buf[RET_STR_LEN] = 0;
-	bip_notify(ic, buf);
+	bip_notify(ic, "%s", buf);
 
 	// TODO: on_connect_send
 
@@ -1394,13 +1413,13 @@ noroom: /* that means the line is larger that RET_STR_LEN. We're not likely to
 			(ch->backlog ? "" : "`"));
 		if (t_written > LINE_SIZE_LIM) {
 			buf[RET_STR_LEN] = 0;
-			bip_notify(ic, buf);
+			bip_notify(ic, "%s", buf);
 			t_written = 0;
 		}
 	}
 noroomchan:
 	buf[RET_STR_LEN] = 0;
-	bip_notify(ic, buf);
+	bip_notify(ic, "%s", buf);
 
 	t_written = snprintf(buf, RET_STR_LEN, "  Status: ");
 	if (t_written >= RET_STR_LEN)
@@ -1460,7 +1479,7 @@ noroomchan:
 	}
 noroomstatus:
 	buf[RET_STR_LEN] = 0;
-	bip_notify(ic, buf);
+	bip_notify(ic, "%s", buf);
 }
 
 void adm_list_all_links(struct link_client *ic)
@@ -1493,7 +1512,7 @@ void adm_list_all_connections(struct link_client *ic)
 
 #define STRORNULL(s) ((s) == NULL ? "unset" : (s))
 
-void adm_info_user(struct link_client *ic, char *name)
+void adm_info_user(struct link_client *ic, const char *name)
 {
 	struct user *u;
 	char buf[RET_STR_LEN + 1];
@@ -1519,8 +1538,7 @@ void adm_info_user(struct link_client *ic, char *name)
 
 noroom:
 	buf[RET_STR_LEN] = 0;
-	bip_notify(ic, buf);
-	t_written = 0;
+	bip_notify(ic, "%s", buf);
 
 #ifdef HAVE_LIBSSL
 	bip_notify(ic, "SSL check mode '%s', stored into '%s'",
@@ -1588,13 +1606,13 @@ void adm_list_users(struct link_client *ic)
 				goto noroom;
 			if (t_written > LINE_SIZE_LIM) {
 				buf[RET_STR_LEN] = 0;
-				bip_notify(ic, buf);
+				bip_notify(ic, "%s", buf);
 				t_written = 0;
 			}
 		}
 noroom:
 		buf[RET_STR_LEN] = 0;
-		bip_notify(ic, buf);
+		bip_notify(ic, "%s", buf);
 	}
 	bip_notify(ic, "-- End of User list");
 }
@@ -1638,13 +1656,13 @@ void adm_list_networks(struct link_client *ic)
 				goto noroom;
 			if (t_written > LINE_SIZE_LIM) {
 				buf[RET_STR_LEN] = 0;
-				bip_notify(ic, buf);
+				bip_notify(ic, "%s", buf);
 				t_written = 0;
 			}
 		}
 noroom:
 		buf[RET_STR_LEN] = 0;
-		bip_notify(ic, buf);
+		bip_notify(ic, "%s", buf);
 	}
 	bip_notify(ic, "-- End of Network list");
 }
@@ -1769,10 +1787,10 @@ int adm_trust(struct link_client *ic, struct line *line)
 		return ERR_PROTOCOL;
 	}
 
-	if (line->elemc != 3)
+	if (irc_line_count(line) != 3)
 		return ERR_PROTOCOL;
 
-	if (!strcasecmp(line->elemv[2], "OK")) {
+	if (irc_line_elem_case_equals(line, 2, "OK")) {
 		/* OK, attempt to trust the cert! */
 		BIO *bio = BIO_new_file(LINK(ic)->user->ssl_check_store, "a+");
 		X509 *trustcert = sk_X509_shift(LINK(ic)->untrusted_certs);
@@ -1787,7 +1805,7 @@ int adm_trust(struct link_client *ic, struct line *line)
 
 		BIO_free_all(bio);
 		X509_free(trustcert);
-	} else if (!strcasecmp(line->elemv[2], "NO")) {
+	} else if (irc_line_elem_case_equals(line, 2, "NO")) {
 		/* NO, discard the cert! */
 		write_line_fast(CONN(ic), ":irc.bip.net NOTICE pouet "
 				":==== Certificate discarded.\r\n");
@@ -1841,11 +1859,17 @@ void bip_notify(struct link_client *ic, char *fmt, ...)
 
 void adm_blreset(struct link_client *ic)
 {
-	log_reinit_all(LINK(ic)->log);
+	log_reset_all(LINK(ic)->log);
 	bip_notify(ic, "backlog resetted for this network.");
 }
 
-void adm_follow_nick(struct link_client *ic, char *val)
+void adm_blreset_store(struct link_client *ic, const char *store)
+{
+	log_reset_store(LINK(ic)->log, store);
+	bip_notify(ic, "backlog resetted for %s.", store);
+}
+
+void adm_follow_nick(struct link_client *ic, const char *val)
 {
 	struct link *link = LINK(ic);
 	if (strcasecmp(val, "TRUE") == 0) {
@@ -1857,7 +1881,7 @@ void adm_follow_nick(struct link_client *ic, char *val)
 	}
 }
 
-void adm_ignore_first_nick(struct link_client *ic, char *val)
+void adm_ignore_first_nick(struct link_client *ic, const char *val)
 {
 	struct link *link = LINK(ic);
 	if (strcasecmp(val, "TRUE") == 0) {
@@ -1875,7 +1899,7 @@ void set_on_connect_send(struct link_client *ic, char *val)
 	char *s;
 
 	if (val != NULL) {
-		list_add_last(&link->on_connect_send, strdup(val));
+		list_add_last(&link->on_connect_send, bip_strdup(val));
 		bip_notify(ic, "added to on_connect_send.");
 	} else {
 		s = list_remove_last(&link->on_connect_send);
@@ -1891,26 +1915,26 @@ void adm_on_connect_send(struct link_client *ic, struct line *line,
 {
 	char buf[ON_CONNECT_MAX_STRSIZE];
 	int t_written = 0;
-	unsigned int i;
+	int i;
 
 	if (!line) {
 		set_on_connect_send(ic, NULL);
 		return;
 	}
 
-	if (line->elemc < 3)
+	if (irc_line_includes(line, 2))
 		return;
 
-	for (i = privmsg + 2; i < line->elemc; i++) {
+	for (i = privmsg + 2; i < irc_line_count(line); i++) {
 		if (t_written) {
 			t_written += snprintf(buf,
 					ON_CONNECT_MAX_STRSIZE - 1 - t_written,
-					" %s", line->elemv[i]);
+					" %s", irc_line_elem(line, i));
 			if (t_written >= ON_CONNECT_MAX_STRSIZE)
 				goto noroom;
 		} else {
 			t_written = snprintf(buf, ON_CONNECT_MAX_STRSIZE - 1,
-					"%s", line->elemv[i]);
+					"%s", irc_line_elem(line, i));
 			if (t_written >= ON_CONNECT_MAX_STRSIZE)
 				goto noroom;
 		}
@@ -1924,7 +1948,7 @@ noroom:
 	goto ok;
 }
 
-void adm_away_nick(struct link_client *ic, char *val)
+void adm_away_nick(struct link_client *ic, const char *val)
 {
 	struct link *link = LINK(ic);
 	if (link->away_nick) {
@@ -1932,30 +1956,35 @@ void adm_away_nick(struct link_client *ic, char *val)
 		link->away_nick = NULL;
 	}
 	if (val != NULL) {
-		link->away_nick = strdup(val);
+		link->away_nick = bip_strdup(val);
 		bip_notify(ic, "away_nick set.");
 	} else {
 		bip_notify(ic, "away_nick cleared.");
 	}
 }
 
-void adm_bip_help(struct link_client *ic, int admin, char *subhelp)
+void adm_bip_help(struct link_client *ic, int admin, const char *subhelp)
 {
 	if (subhelp == NULL) {
 		if (admin) {
-			bip_notify(ic, "/BIP RELOAD # Re-read bip configuration "
-				"and apply changes. /!\\ VERY UNSTABLE !");
-			bip_notify(ic, "/BIP INFO user <username> # show a user's "
-				"configuration");
-			bip_notify(ic, "/BIP LIST networks|users|connections|all_links"
-				"|all_connections");
-			bip_notify(ic, "/BIP ADD_CONN <connection name> <network>");
+			bip_notify(ic, "/BIP RELOAD # Re-read bip "
+					"configuration and apply changes.");
+			bip_notify(ic, "/BIP INFO user <username> "
+					"# show a user's configuration");
+			bip_notify(ic, "/BIP LIST networks|users|connections|"
+					"all_links|all_connections");
+			bip_notify(ic, "/BIP ADD_CONN <connection name> "
+					"<network>");
 			bip_notify(ic, "/BIP DEL_CONN <connection name>");
 		} else {
 			bip_notify(ic, "/BIP LIST networks|connections");
 		}
-		bip_notify(ic, "/BIP JUMP # jump to next server (in same network)");
-		bip_notify(ic, "/BIP BLRESET # reset backlog (this connection only). Add -q flag and the operation is quiet.");
+		bip_notify(ic, "/BIP JUMP # jump to next server (in same "
+				"network)");
+		bip_notify(ic, "/BIP BLRESET [channel|query]# reset backlog "
+				"(this connection only). Add -q flag and the "
+				"operation is quiet. You can specify a channel "
+				"or a nick to reset only this channel/query.");
 #ifdef HAVE_LIBSSL
 		bip_notify(ic, "/BIP TRUST # trust this server certificate");
 #endif
@@ -1967,6 +1996,8 @@ void adm_bip_help(struct link_client *ic, int admin, char *subhelp)
 		bip_notify(ic, "/BIP ON_CONNECT_SEND # Clears on_connect_send");
 		bip_notify(ic, "/BIP AWAY_NICK <nick> # Set away nick");
 		bip_notify(ic, "/BIP AWAY_NICK # clear away nick");
+		bip_notify(ic, "/BIP BACKLOG [n] # backlog text of the n last "
+				"hours");
 	} else if (admin && strcasecmp(subhelp, "RELOAD") == 0) {
 		bip_notify(ic, "/BIP RELOAD (admin only) :");
 		bip_notify(ic, "  Reloads bip configuration file and apply "
@@ -2050,20 +2081,23 @@ void adm_bip_help(struct link_client *ic, int admin, char *subhelp)
 	}
 }
 
-int adm_bip(bip_t *bip, struct link_client *ic, struct line *line,
-		unsigned int privmsg)
+int adm_bip(bip_t *bip, struct link_client *ic, struct line *line, int privmsg)
 {
 	int admin = LINK(ic)->user->admin;
 
 	if (privmsg) {
-		char *linestr = line->elemv[2];
-		char *ptr = line->elemv[2], *eptr;
+		char *linestr, *elemstr;
+		char *ptr, *eptr;
 		int slen;
 
-		if (line->elemc != 3)
+		if (irc_line_count(line) != 3)
 			return OK_FORGET;
 
-		line->elemc--;
+		linestr = irc_line_pop(line);
+		ptr = linestr;
+
+		/* all elem size <= linestr size */
+		elemstr = bip_malloc(strlen(linestr) + 1);
 
 		while((eptr = strstr(ptr, " "))) {
 			slen = eptr - ptr;
@@ -2071,34 +2105,28 @@ int adm_bip(bip_t *bip, struct link_client *ic, struct line *line,
 				ptr++;
 				continue;
 			}
-			line->elemv = realloc(line->elemv,
-					(line->elemc + 1) * sizeof(char *));
-			line->elemv[line->elemc] = malloc(slen + 1);
-			memcpy(line->elemv[line->elemc], ptr, slen);
-			line->elemv[line->elemc][slen] = 0;
-			line->elemc++;
+			memcpy(elemstr, ptr, slen);
+			elemstr[slen] = 0;
+			irc_line_append(line, elemstr);
 			ptr = eptr + 1;
 		}
 		eptr = ptr + strlen(ptr);
 		slen = eptr - ptr;
 		if (slen != 0) {
-			line->elemv = realloc(line->elemv,
-				      (line->elemc + 1) * sizeof(char *));
-			slen = eptr - ptr;
-			line->elemv[line->elemc] = malloc(slen + 1);
-			memcpy(line->elemv[line->elemc], ptr, slen);
-			line->elemv[line->elemc][slen] = 0;
-			line->elemc++;
+			memcpy(elemstr, ptr, slen);
+			elemstr[slen] = 0;
+			irc_line_append(line, elemstr);
 		}
+		free(elemstr);
 		free(linestr);
 	}
 
-	if (line->elemc < privmsg + 2)
+	if (!irc_line_includes(line, privmsg + 1))
 		return OK_FORGET;
 
-	mylog(LOG_INFO, "/BIP %s from %s", line->elemv[privmsg + 1],
+	mylog(LOG_INFO, "/BIP %s from %s", irc_line_elem(line, privmsg + 1),
 			LINK(ic)->user->name);
-	if (strcasecmp(line->elemv[privmsg + 1], "RELOAD") == 0) {
+	if (strcasecmp(irc_line_elem(line, privmsg + 1), "RELOAD") == 0) {
 		if (!admin) {
 			bip_notify(ic, "-- You're not allowed to reload bip");
 			return OK_FORGET;
@@ -2106,132 +2134,148 @@ int adm_bip(bip_t *bip, struct link_client *ic, struct line *line,
 		bip_notify(ic, "-- Reloading bip...");
 		bip->reloading_client = ic;
 		sighup = 1;
-	} else if (strcasecmp(line->elemv[privmsg + 1], "LIST") == 0) {
-		if (line->elemc != privmsg + 3) {
+	} else if (strcasecmp(irc_line_elem(line, privmsg + 1), "LIST") == 0) {
+		if (irc_line_count(line) != privmsg + 3) {
 			bip_notify(ic, "-- LIST command needs one argument");
 			return OK_FORGET;
 		}
 
-		if (admin && strcasecmp(line->elemv[privmsg + 2],
+		if (admin && strcasecmp(irc_line_elem(line, privmsg + 2),
 					"users") == 0) {
 			adm_list_users(ic);
-		} else if (strcasecmp(line->elemv[privmsg + 2],
+		} else if (strcasecmp(irc_line_elem(line, privmsg + 2),
 					"networks") == 0) {
 			adm_list_networks(ic);
-		} else if (strcasecmp(line->elemv[privmsg + 2],
+		} else if (strcasecmp(irc_line_elem(line, privmsg + 2),
 					"connections") == 0) {
 			adm_list_connections(ic, NULL);
-		} else if (admin && strcasecmp(line->elemv[privmsg + 2],
+		} else if (admin && strcasecmp(irc_line_elem(line, privmsg + 2),
 					"all_connections") == 0) {
 			adm_list_all_connections(ic);
-		} else if (admin && strcasecmp(line->elemv[privmsg + 2],
+		} else if (admin && strcasecmp(irc_line_elem(line, privmsg + 2),
 					"all_links") == 0) {
 			adm_list_all_links(ic);
 		} else {
 			bip_notify(ic, "-- Invalid LIST request");
 		}
-	} else if (strcasecmp(line->elemv[privmsg + 1], "INFO") == 0) {
-		if (line->elemc < privmsg + 3) {
-			bip_notify(ic, "-- INFO command needs at least one argument");
+	} else if (strcasecmp(irc_line_elem(line, privmsg + 1), "INFO") == 0) {
+		if (!irc_line_includes(line, privmsg + 2)) {
+			bip_notify(ic, "-- INFO command needs at least one "
+					"argument");
 			return OK_FORGET;
 		}
 
-		if (admin && strcasecmp(line->elemv[privmsg + 2],
+		if (admin && irc_line_elem_case_equals(line, privmsg + 3,
 					"user") == 0) {
-			if (line->elemc == privmsg + 4) {
-				adm_info_user(ic, line->elemv[privmsg + 3]);
+			if (irc_line_count(line) == privmsg + 4) {
+				adm_info_user(ic,
+					irc_line_elem(line, privmsg + 3));
 			} else {
 				bip_notify(ic, "-- INFO USER command needs one"
 					" argument");
 			}
 #if 0
-			TODO
-		} else if (strcasecmp(line->elemv[privmsg + 2],
-				      "network") == 0) {
-			if (line->elemc == privmsg + 4) {
-				adm_info_network(ic, line->elemv[privmsg + 3]);
-			} else {
-				bip_notify(ic, "/BIP INFO network needs one "
-					"argument");
-			}
+			TODO network info
 #endif
 		} else {
 			bip_notify(ic, "-- Invalid INFO request");
 		}
-	} else if (strcasecmp(line->elemv[privmsg + 1], "JUMP") == 0) {
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "JUMP")) {
 		if (LINK(ic)->l_server) {
 			WRITE_LINE1(CONN(LINK(ic)->l_server), NULL, "QUIT",
 					"jumpin' jumpin'");
 			connection_close(CONN(LINK(ic)->l_server));
 		}
 		bip_notify(ic, "-- Jumping to next server");
-	} else if (strcasecmp(line->elemv[privmsg + 1], "BLRESET") == 0) {
-		if (line->elemc == privmsg + 3 &&
-				strcmp(line->elemv[privmsg + 2], "-q") == 0) {
-			log_reinit_all(LINK(ic)->log);
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "BLRESET")) {
+		if (irc_line_includes(line, privmsg + 2)) {
+			if (irc_line_elem_equals(line, privmsg + 2, "-q")) {
+				if (irc_line_includes(line, privmsg + 3)) {
+					log_reset_store(LINK(ic)->log,
+						irc_line_elem(line,
+							privmsg + 3));
+				} else {
+					log_reset_all(LINK(ic)->log);
+				}
+			} else {
+				adm_blreset_store(ic, irc_line_elem(line,
+							privmsg + 2));
+			}
 		} else {
 			adm_blreset(ic);
 		}
-	} else if (strcasecmp(line->elemv[privmsg + 1], "HELP") == 0) {
-		if (line->elemc == privmsg + 2)
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "HELP")) {
+		if (irc_line_count(line) == privmsg + 2)
 			adm_bip_help(ic, admin, NULL);
-		else if (line->elemc == privmsg + 3)
-			adm_bip_help(ic, admin, line->elemv[privmsg + 2]);
+		else if (irc_line_count(line) == privmsg + 3)
+			adm_bip_help(ic, admin,
+					irc_line_elem(line, privmsg + 2));
 		else
 			bip_notify(ic,
 				"-- HELP command needs at most one argument");
-	} else if (strcasecmp(line->elemv[privmsg + 1], "FOLLOW_NICK") == 0) {
-		if (line->elemc != privmsg + 3) {
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "FOLLOW_NICK")) {
+		if (irc_line_count(line) != privmsg + 3) {
 			bip_notify(ic,
 				"-- FOLLOW_NICK command needs one argument");
 			return OK_FORGET;
 		}
-		adm_follow_nick(ic, line->elemv[privmsg + 2]);
-	} else if (strcasecmp(line->elemv[privmsg + 1],
-				"IGNORE_FIRST_NICK") == 0) {
-		if (line->elemc != privmsg + 3) {
+		adm_follow_nick(ic, irc_line_elem(line, privmsg + 2));
+	} else if (irc_line_elem_case_equals(line, privmsg + 1,
+				"IGNORE_FIRST_NICK")) {
+		if (irc_line_count(line) != privmsg + 3) {
 			bip_notify(ic, "-- IGNORE_FIRST_NICK "
 				"command needs one argument");
 			return OK_FORGET;
 		}
-		adm_ignore_first_nick(ic, line->elemv[privmsg + 2]);
-	} else if (strcasecmp(line->elemv[privmsg + 1],
-				"ON_CONNECT_SEND") == 0) {
-		if (line->elemc == privmsg + 2) {
+		adm_ignore_first_nick(ic, irc_line_elem(line, privmsg + 2));
+	} else if (irc_line_elem_case_equals(line, privmsg + 1,
+				"ON_CONNECT_SEND")) {
+		if (irc_line_count(line) == privmsg + 2) {
 			adm_on_connect_send(ic, NULL, 0);
-		} else if (line->elemc >= privmsg + 3) {
+		} else if (irc_line_includes(line, privmsg + 2)) {
 			adm_on_connect_send(ic, line, privmsg);
 		} else {
 			bip_notify(ic, "-- ON_CONNECT_SEND command needs at "
 				"least one argument");
 		}
-	} else if (strcasecmp(line->elemv[privmsg + 1], "AWAY_NICK") == 0) {
-		if (line->elemc == privmsg + 2) {
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "AWAY_NICK")) {
+		if (irc_line_count(line) == privmsg + 2) {
 			adm_away_nick(ic, NULL);
-		} else if (line->elemc == privmsg + 3) {
-			adm_away_nick(ic, line->elemv[privmsg + 2]);
+		} else if (irc_line_count(line) == privmsg + 3) {
+			adm_away_nick(ic, irc_line_elem(line, privmsg + 2));
 		} else {
 			bip_notify(ic, "-- AWAY_NICK command needs zero or one"
 				" argument");
 		}
-	} else if (admin &&
-			strcasecmp(line->elemv[privmsg + 1], "ADD_CONN") == 0) {
-		if (line->elemc != privmsg + 4) {
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "BACKLOG")) {
+		if (irc_line_count(line) == privmsg + 2) {
+			irc_cli_backlog(ic, 0);
+		} else if (irc_line_count(line) == privmsg + 3) {
+			int hours = atoi(irc_line_elem(line, privmsg + 2));
+			irc_cli_backlog(ic, hours);
+		} else {
+			bip_notify(ic, "-- BACKLOG takes 0 or one argument");
+		}
+	} else if (admin && irc_line_elem_case_equals(line, privmsg + 1,
+				"ADD_CONN")) {
+		if (irc_line_count(line) != privmsg + 4) {
 			bip_notify(ic, "/BIP ADD_CONN <connection name> "
 					"<network name>");
 		} else {
-			adm_bip_addconn(bip, ic, line->elemv[privmsg + 2],
-					line->elemv[privmsg + 3]);
+			adm_bip_addconn(bip, ic,
+					irc_line_elem(line, privmsg + 2),
+					irc_line_elem(line, privmsg + 3));
 		}
-	} else if (admin &&
-			strcasecmp(line->elemv[privmsg + 1], "DEL_CONN") == 0) {
-		if (line->elemc != privmsg + 3) {
+	} else if (admin && irc_line_elem_case_equals(line, privmsg + 1,
+				"DEL_CONN")) {
+		if (irc_line_count(line) != privmsg + 3) {
 			bip_notify(ic, "/BIP DEL_CONN <connection name>");
 		} else {
-			adm_bip_delconn(bip, ic, line->elemv[privmsg + 2]);
+			adm_bip_delconn(bip, ic,
+					irc_line_elem(line, privmsg + 2));
 		}
 #ifdef HAVE_LIBSSL
-	} else if (strcasecmp(line->elemv[privmsg + 1], "TRUST") == 0) {
+	} else if (irc_line_elem_case_equals(line, privmsg + 1, "TRUST")) {
 		return adm_trust(ic, line);
 #endif
 	} else {
