@@ -334,7 +334,7 @@ void write_line_fast(connection_t *cn, char *line)
 	r = write_socket(cn, line);
 	switch (r) {
 	case WRITE_KEEP:
-		list_add_first(cn->outgoing, strdup(line));
+		list_add_first(cn->outgoing, bip_strdup(line));
 		break;
 	case WRITE_ERROR:
 		cn->connected = CONN_ERROR;
@@ -347,9 +347,14 @@ void write_line_fast(connection_t *cn, char *line)
 	}
 }
 
+void write_lines(connection_t *cn, list_t *lines)
+{
+	list_append(cn->outgoing, lines);
+}
+
 void write_line(connection_t *cn, char *line)
 {
-	list_add_last(cn->outgoing, strdup(line));
+	list_add_last(cn->outgoing, bip_strdup(line));
 }
 
 list_t *read_lines(connection_t *cn, int *error)
@@ -479,7 +484,7 @@ static void data_find_lines(connection_t *cn)
 		if (ssz >= 1) {
 			if (p[len - 1] == '\r')
 				ssz--;
-			buf = malloc(ssz + 1);
+			buf = bip_malloc(ssz + 1);
 			memcpy(buf, p + lastlen, ssz);
 			buf[ssz] = 0;
 
@@ -817,8 +822,11 @@ list_t *wait_event(list_t *cn_list, int *msec, int *nc)
 
 	/* if no connection is active, return the list... empty... */
 	if (maxfd == -1) {
-		usleep(*msec * 1000);
-		*msec = 0;
+		struct timespec req, rem;
+		req.tv_sec = *msec * 1000;
+		req.tv_nsec = 0;
+		nanosleep(&req, &rem);
+		*msec = rem.tv_sec;
 		return cn_newdata;
 	}
 
@@ -891,9 +899,7 @@ static void create_socket(char *dsthostname, char *dstport, char *srchostname,
 
 	cn->connected = CONN_ERROR;
 	cdata = (struct connecting_data *)
-		malloc(sizeof(struct connecting_data));
-	if (!cdata)
-		fatal("Out of memory.");
+		bip_malloc(sizeof(struct connecting_data));
 	cdata->dst = cdata->src = cdata->cur = NULL;
 
 	err = getaddrinfo(dsthostname, dstport, &hint, &cdata->dst);
@@ -997,8 +1003,8 @@ static connection_t *connection_init(int anti_flood, int ssl, int timeout,
 	char *incoming;
 	list_t *outgoing;
 
-	conn = (connection_t *)calloc(sizeof(connection_t), 1);
-	incoming = (char*)malloc(sizeof(char) * CONN_BUFFER_SIZE);
+	conn = (connection_t *)bip_calloc(sizeof(connection_t), 1);
+	incoming = (char *)bip_malloc(CONN_BUFFER_SIZE);
 	outgoing = list_new(NULL);
 
 	conn->anti_flood = anti_flood;
@@ -1027,6 +1033,73 @@ static connection_t *connection_init(int anti_flood, int ssl, int timeout,
 	return conn;
 }
 
+#ifdef HAVE_LIBSSL
+
+#include "moduli.h"
+
+/* from postfix tls impl */
+static DH *dh_512(void)
+{
+	DH *dh;
+	static DH *dh_512;
+
+	if (dh_512 == NULL) {
+		if ((dh = DH_new()) == NULL) {
+			mylog(LOG_WARN, "SSL: cannot create DH parameter set");
+			return (0);
+		}
+		dh->p = BN_bin2bn(dh512_p, sizeof(dh512_p), (BIGNUM *) 0);
+		dh->g = BN_bin2bn(dh512_g, sizeof(dh512_g), (BIGNUM *) 0);
+		if ((dh->p == NULL) || (dh->g == NULL)) {
+			mylog(LOG_WARN, "SSL: cannot load compiled-in DH "
+					"parameters");
+			DH_free(dh);
+			return (0);
+		} else
+			dh_512 = dh;
+	}
+	return dh_512;
+}
+
+/* tls_get_dh_1024 - get 1024-bit DH parameters, compiled-in or from file */
+static DH *dh_1024(void)
+{
+	DH *dh;
+	static DH *dh_1024;
+
+	if (dh_1024 == NULL) {
+		if ((dh = DH_new()) == NULL) {
+			mylog(LOG_WARN, "SSL: cannot create DH parameter set");
+			return (0);
+		}
+		dh->p = BN_bin2bn(dh1024_p, sizeof(dh1024_p), (BIGNUM *) 0);
+		dh->g = BN_bin2bn(dh1024_g, sizeof(dh1024_g), (BIGNUM *) 0);
+		if ((dh->p == NULL) || (dh->g == NULL)) {
+			mylog(LOG_WARN, "SSL: cannot load compiled-in DH "
+					"parameters");
+			DH_free(dh);
+			return (0);
+		} else
+			dh_1024 = dh;
+	}
+	return (dh_1024);
+}
+
+/* ripped from postfix's tls_dh.c */
+static DH *tmp_dh_cb(SSL *ssl_unused, int export, int keylength)
+{
+	DH *ret;
+
+	if (export && keylength == 512) {
+		mylog(LOG_WARN, "SSL: using 512 bits diffie hellman moduli");
+		ret = dh_512();
+	} else {
+		ret = dh_1024();
+	}
+	return ret;
+}
+#endif
+
 connection_t *accept_new(connection_t *cn)
 {
 	connection_t *conn;
@@ -1067,6 +1140,10 @@ connection_t *accept_new(connection_t *cn)
 						conf_ssl_certfile,
 						SSL_FILETYPE_PEM))
 				mylog(LOG_WARN, "SSL: Unable to load key file");
+
+			/* diffie hellman key generation need us to feed some
+			   data that can be static ... */
+			SSL_CTX_set_tmp_dh_callback(sslctx, tmp_dh_cb);
 		}
 
 		conn->ssl_h = SSL_new(sslctx);
@@ -1586,9 +1663,7 @@ static char *socket_ip(int fd, int remote)
 		return NULL;
 	}
 
-	ip = malloc(65);
-	if (ip == NULL)
-		fatal("malloc");
+	ip = bip_malloc(65);
 
 	switch (addr.sa_family) {
 	case AF_INET:
